----------------------------------------------------------------------------------
-- Company: The Hong Kong Polytechnic University
-- Engineer: Alexandr Melnikov
-- 
-- Create Date:    12:14:44 08/03/2016 
-- Design Name: 
-- Module Name:    dig2an - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created

-- Additional Comments: 
-- DAC block, handles conversion of the data received from DTE 
-- to analog form
-- DAC uses SPI (3-wire) protocol.
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity dig2an is
    Port ( clk100mhz : in  STD_LOGIC;
			  din_dac : in STD_LOGIC_VECTOR (11 downto 0); -- data to write to DAC (from FuncGen)
			  baud_clk : in STD_LOGIC; -- Baud tick to synchronize DAC block with FuncGen
           CS : out  STD_LOGIC;   -- chip select (LOW when write DAC)
           MOSI : out  STD_LOGIC; -- DAC's data line
			  SCLK : out  STD_LOGIC; -- DAC's clock line
			  st_wrt : in STD_LOGIC;    -- start write DAC (from FuncGen)
			  wrt_done : out STD_LOGIC);-- write done (to FuncGen)
end dig2an;

architecture Behavioral of dig2an is
-- DAC state machine
type state_type is (idle_dac, init_dac, func_dac, wrt_dac, pause);
signal state : state_type := idle_dac;
-- DAC signals
-- counter to track DAC's output
signal count : integer range 0 to 32 := 0; -- to control DAC operation
-- data to write to DAC
signal data : std_logic_vector (31 downto 0) := x"00000000"; 
-- command word to set up DAC, internal reference is used
signal setup_dac : std_logic_vector (31 downto 0) := x"08000001";
-- flag indicating initialization of the DAC 
signal dac_init : std_logic := '0';

-- local clock signals
signal clkdiv : integer range 0 to 49 := 0; -- clock divider for 1MHz DAC clock
signal clk1mhz : std_logic := '0'; -- 1MHz DAC clock


begin
	
	-- this process takes care of DAC clock generation
	clock_dac: process (clk100mhz) 
	begin
		if(rising_edge(clk100mhz)) then -- clkdiv_max=1 for 50MHz clock
			if (clkdiv=49)then
				clkdiv <= 0;
				clk1mhz <= not(clk1mhz);
			else
				clkdiv <= clkdiv+1;
			end if;
		end if;
	end process clock_dac;
	
	
	dac_main: process (clk1mhz, baud_clk, st_wrt)
	
	begin
		if (st_wrt='0') then -- st_write acts as internal Reset for DAC module
			state <= idle_dac;
			CS <= '1';
			MOSI <= '0';
			count <= 0;
			wrt_done <= '0';
		elsif (rising_edge(clk1mhz)) then
			case state is
			
				-- DAC stays idle until the upper level module (FuncGen)
				-- tells it to start writing the DAC
				when idle_dac =>
					CS <= '1';
					MOSI <= '0';
					count <= 0;
					wrt_done <= '0';
					-- if DAC hasn't been initialized, do it first (it's done once only)
					if (dac_init = '0') then
						CS <= '1';
						state <= init_dac;
					else -- if initialization has been done, start writing it
						CS <= '1';
						state <= func_dac;
					end if;
				
				-- in this state initialization word is being written to DAC
				when init_dac =>
					if (count <= 31) then
						CS <= '0'; -- CS is set LOW when DAC is being written
						wrt_done <= '0';
						MOSI <= setup_dac(31-count);
						count <= count+1;
						state <= init_dac;
					elsif (count = 32) then
						CS <= '1';
						wrt_done <= '0';
						count <= 0;
						dac_init <= '1'; -- set the flag indicating that initialization is done
						MOSI <= '0';
						state <= func_dac;
					end if;
				
				-- convert data to be written to DAC (received from FuncGen) 
				-- into DAC's command word (see datasheet)
				when func_dac =>
					wrt_done <= '0';
					CS <= '1';
					MOSI <= '0';
					data <= x"030" & din_dac & x"00";					
					state <= wrt_dac;
				
				-- write DAC's command word to DAC
				when wrt_dac =>
					if (count <= 31) then
						CS <= '0';
						wrt_done <= '0';
						MOSI <= data(31-count);
						count <= count+1;
						state <= wrt_dac;
					elsif (count = 32) then
						CS <= '1';
						wrt_done <= '1'; -- set the flag indicating that write is done
						count <= 0;
						MOSI <= '0';
						state <= pause;
					end if;
				
				-- stays in this state until the next Baud tick to sinchronize
				-- DAC's module operation with upper level module (FuncGen)
				when pause =>
					if (baud_clk = '1') then
						wrt_done <= '0'; -- clear write flag
						state <= idle_dac;
					else
						wrt_done <= '1';
						state <= pause;
					end if;
				
			end case;
		end if;
	end process dac_main;
	
SCLK <= clk1mhz; -- drive DAC with the locally generated 1MHz clock signal 
	
end Behavioral;

