----------------------------------------------------------------------------------
-- Company: The Hong Kong Polytechnic University
-- Engineer: Alexandr Melnikov
-- 
-- Create Date:    16:35:44 08/24/2016 
-- Design Name: 
-- Module Name:    func_gen - Behavioral 
-- Project Name: 
-- Target Devices: Nexys4 DDR
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created

-- Additional Comments: 
-- This projects implements a custom function generator (FuncGen).
-- Command signal to the function generator is supplied from Matlab through
-- on-board UART bridge as a 16-bit long command word (unsigned integer).
-- Digital command signal is converted into corresponding voltage signal by DAC (Pmod DA4),
-- which can be used to drive external device.
-- Feedback, implemented on the ADC (PmodAD2), allows user to control the
-- actual level of the voltage signal. The feedback signal is sent back to the DTE (PC, Matlab),
-- using the same UART bridge.
-- The current level of the voltage feedback signal is displayed on the on-board
-- 8-digit seven segment display.
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

library ieee_proposed;
use ieee_proposed.fixed_pkg.all;
--use ieee.fixed_generic_pkg.all;

entity func_gen is
-- 8 data bits per word, 1 stop bit, 0 parity bits, 2 words = 20 bits
-- Baud Rates and Divisors for 16*oversampling mode:
	-- error << 1%
-- 9600 b/s -> 651;
-- 14400 b/s -> 434;
	-- error < 1%
-- 19200 b/s -> 326;
-- 38400 b/s -> 163;
	-- error ~ 0.5%
-- 57600 b/s -> 109;
-- 115200 b/s -> 54;
-- 230400 b/s -> 27;
	-- error ~ 3%;
-- 460800 b/s -> 14;
-- 921600 b/s -> 7; -- not recommended for 100mhz clk (error is too high)
-- Baud Rates and Divisors for 13*oversampling mode:
	-- error << 1%
-- 9600 b/s -> 801;
-- 14400 b/s -> 534;
-- 19200 b/s -> 401;
	-- error < 1%
-- 38400 b/s -> 200;
	-- error ~ 0.33%
-- 57600 b/s -> 134;
-- 115200 b/s -> 67;
	-- error > 1%;
-- 230400 b/s -> 33;
-- 460800 b/s -> 17;
	-- error > 4%;
-- 921600 b/s -> 8; -- not recommended for 100mhz clk (error is too high)

	 generic ( N_bit : integer := 20;  -- total lenght of data words
				  OSR : integer := 13;    -- oversampling rate
				  BRD : integer := 17;   -- baud rate divisor
				  M_clk : integer := 49999; -- display clock divisor
				  Mi : integer := 4;        -- num of bits for integer part
				  Mf : integer := 7;        -- num of bits for fractional part
				  res : ufixed (0 downto -16) :="00000000000101000"); -- DAC/ADC resolution

    Port ( clk100mhz : in  STD_LOGIC;
			  -- uart inputs/outputs
           uart_cts : out  STD_LOGIC;
           uart_rts : in  STD_LOGIC;
			  uart_txd_in : in  STD_LOGIC;
           uart_rxd_out : out  STD_LOGIC;
			  -- data converters inputs/outputs
			  -- dac
			  CS : out  STD_LOGIC;
			  MOSI : out  STD_LOGIC;
			  SCLK : out  STD_LOGIC;
			  -- adc
			  SDA : inout  STD_LOGIC;
			  SCL : out  STD_LOGIC;
			  -- display inputs/outputs
			  an : out STD_LOGIC_VECTOR (7 downto 0); -- common anodes LOW
			  ca : out std_logic; -- top
			  cb : out std_logic; -- top-right
			  cc : out std_logic; -- bottom-right
			  cd : out std_logic; -- bottom
			  ce : out std_logic; -- bottom-left
			  cf : out std_logic; -- top-left
			  cg : out std_logic; -- middle
			  dp : out std_logic -- decimal point
			  );
end func_gen;

architecture Behavioral of func_gen is
-- FSM to control FuncGen
type state_type is (idle, rd_uart, wrt_dac, rd_adc, func, wrt_uart);
signal state : state_type := idle;
-- Baud tick
signal baud_clk : std_logic;
-- Copy of the command signal received from Matlab via UART Rx
signal dout_rx : std_logic_vector (11 downto 0):= (others => '0');
-- Copy of the feedback signal received from ADC
signal dout_adc : std_logic_vector (11 downto 0):= (others => '0');
-- Copy of the feedback signal to be sent via UART Tx
signal din_tx : std_logic_vector (N_bit-1 downto 0):= (others => '0');
-- Control signals for FSM
signal start_write_uart : std_logic := '0';
signal write_done_uart : std_logic;
signal start_read_uart : std_logic := '0';
signal read_done_uart : std_logic;
signal start_write_dac : std_logic := '0';
signal write_done_dac : std_logic;
signal start_read_adc : std_logic := '0';
signal read_done_adc : std_logic;

-- Instantiation of the FuncGen components
-- Baud Rate Generator
component brgen 
					 generic (M : integer); 			 -- Baud Rate Divisor
					 port ( clk100mhz : in std_logic; -- board clock
							  bclk : out std_logic); 	 -- Baud tick
end component;
-- UART Tx
component tx 
				 generic ( N_bit : integer;    -- length of the data word (bits)
							  N_sampl : integer); -- oversampling rate
				 port ( clk100mhz : in std_logic;     -- board clock
						  uart_rxd_out : out std_logic; -- output Tx port
						  din : in std_logic_vector (N_bit-1 downto 0); -- output (feedback) signal
						  b_tick : in std_logic;    	  -- Baud tick
						  wrt_done : out std_logic; 	  -- inner control signal (indicates end of transmission)
						  st_write : in std_logic); 	  -- inner control signal (starts transmission)
end component;
-- UART Rx
component rx 
				 generic ( N_bit : integer;	-- length of the data word (bits)
							  N_sampl : integer);-- oversampling rate
				 port ( clk100mhz : in  STD_LOGIC;	 -- board clock
						  uart_txd_in : in  STD_LOGIC; -- input Rx port
						  dout : out std_logic_vector (11 downto 0);-- input command signal
						  b_tick : in std_logic;		-- Baud tick
						  rd_done : out std_logic;		-- inner control signal (indicates end of reception)
						  st_read : in std_logic);		-- inner control signal (starts reception)
end component;
-- DAC
component dig2an port ( clk100mhz : in  STD_LOGIC; -- board clock
								din_dac : in STD_LOGIC_VECTOR (11 downto 0); -- data to write to dac (command signal)
								baud_clk : in STD_LOGIC;	-- Baud tick
								st_wrt : in STD_LOGIC;		-- inner control signal (starts conversion)
								wrt_done : out STD_LOGIC;  -- inner control signal (indicates end of conversion)
								-- DAC signals, see datasheet
								CS : out  STD_LOGIC; 	-- chip select
								MOSI : out  STD_LOGIC;  -- data line
								SCLK : out  STD_LOGIC); -- clock line
								
end component;
-- ADC
component a2d port ( clk100mhz : in  STD_LOGIC;		-- board clock
							dout : out  STD_LOGIC_VECTOR (11 downto 0); -- data read by ADC
							start_read : in  STD_LOGIC;	-- inner control signal (starts conversion)
							read_done : out STD_LOGIC;	 	-- inner control signal (indicates end of conversion)
							-- ADC signals, see datasheet
							SDA : inout  STD_LOGIC; 		-- data line
							SCL : out  STD_LOGIC); 			-- clock line
							
end component;
-- 8-digit SSD
component disp
				generic ( M_clk : integer; -- display clock divisor
							 Mi : integer;    -- number of bits for integer part
							 Mf : integer;    -- number of bits for fractional part
							 res : ufixed (0 downto -16)); -- DAC/ADC resolution
				port ( clk100mhz : in  STD_LOGIC; -- board clock
						 data2disp : in STD_LOGIC_VECTOR (11 downto 0); -- data to display (ADC's readings)
						 -- SSD control signals
						 an : out STD_LOGIC_VECTOR (7 downto 0);
						 ca : out std_logic;
					    cb : out std_logic;
					    cc : out std_logic;
					    cd : out std_logic;
					    ce : out std_logic;
					    cf : out std_logic;
					    cg : out std_logic;
					    dp : out std_logic);
end component;

begin
-- Connection between formal and actual signals
BaudRateGen: brgen 
						 generic map ( M => BRD) -- pass Baud Rate Divisor value to Baud Rate Generator
						 port map ( clk100mhz => clk100mhz, -- connect clock input to Baud Rate Generator
										bclk => baud_clk); -- return Baud tick signal to FuncGen
								
Transmitter: tx 
					 generic map ( N_bit => N_bit, -- pass word length and
									   N_sampl => OSR) -- oversampling rate to Tx module
					 port map ( clk100mhz => clk100mhz, 		-- connect clock input to Tx module
								   uart_rxd_out => uart_rxd_out, -- connect output UART Rx signal to Tx module
									din => din_tx, 					-- pass feedback signal to Tx module
									b_tick => baud_clk, 				-- pass Baud tick to Tx module
									wrt_done => write_done_uart,  -- return control signal from Tx module
									st_write => start_write_uart);-- pass control signal to Tx module
									
Receiver: rx 
				 generic map ( N_bit => N_bit, -- pass word length and
									N_sampl => OSR) -- oversampling rate to Rx module
				 port map ( clk100mhz => clk100mhz, 	 -- connect clock input to Rx module
								uart_txd_in => uart_txd_in, -- connect UART Tx input to Rx module
								dout => dout_rx,				 -- return command signal from Rx module
								b_tick => baud_clk,			 -- pass Baud tick to Rx module
								rd_done => read_done_uart,  -- return control signal from Rx module
								st_read => start_read_uart);-- pass control signal to Rx module
								
DAC: dig2an port map ( clk100mhz => clk100mhz, -- connect clock input to DAC module
							  CS => CS,					  -- connect SC (chip select) output to DAC module
							  MOSI => MOSI,			  -- connect MOSI (data line) output to DAC module
							  SCLK => SCLK,			  -- connect SCLK (clock line) output to DAC module
							  din_dac => dout_rx,     -- pass received by Rx data to DAC module
							  baud_clk => baud_clk,	  -- pass Baud tick to DAC module
							  st_wrt => start_write_dac,  -- pass control signal to DAC module
							  wrt_done => write_done_dac);-- return control signal from DAC module

ADC : a2d port map ( clk100mhz => clk100mhz,			-- connect clock input to ADC module
							dout => dout_adc,					-- receive conversion results (feedback) from ADC module
							start_read => start_read_adc, -- pass control signal to ADC module
							read_done => read_done_adc,   -- return control signal from ADC module
							SDA => SDA,   -- connect SDA (data line) input/output to ADC module
							SCL => SCL);  -- connect SCL(clock line) output to ADC module
							
Display: disp
					generic map ( M_clk => M_clk, -- pass display clock's divisor value to Diplay module
									  Mi => Mi,			-- pass number of bits for integer part value to Diplay module
									  Mf => Mf,			-- pass number of bits for fractional part value to Diplay module
									  res => res)     -- pass resolution value to Diplay module
					port map ( clk100mhz => clk100mhz, -- connect clock input to Display module
								  data2disp => dout_adc,  -- pass feedback signal to Display module
								  -- connect SSD outputs to Display module 
								  an => an,
								  ca => ca,
								  cb => cb,
								  cc => cc,
								  cd => cd,
								  ce => ce,
								  cf => cf,
								  cg => cg,
								  dp => dp);
	
	fg: process (clk100mhz, uart_rts, baud_clk)
	begin

		if(rising_edge(clk100mhz) and baud_clk = '1')then
			case state is
			-- in the idle state FuncGen waits for the DTE to assert UART_RTS LOW
			-- which indicated the data to be transmitted from DTE to the board
			when idle =>
				uart_cts <= '1';
				start_write_uart <= '0';
				start_write_dac <= '0';
				start_read_adc <= '0';
				if(uart_rts='1') then
					start_read_uart <= '0';
					state<=idle;
				else
					start_read_uart <= '1';
					-- set UART_CTS low, to indicate that the board 
					-- is ready to receive from DTE (PC, Matlab)
					uart_cts <= '0'; 
					state <= rd_uart;
				end if;
			
			-- in this state FuncGen reads the signal transmitted
			-- from DTE
			when rd_uart =>
				uart_cts <= '0';
				start_write_uart <= '0';
				start_write_dac <= '0';
				start_read_uart <= '1';
				start_read_adc <= '0';
				-- Rx block keeps reading until read_done HIGH is asserted
				if (read_done_uart = '1') then 
					-- stop reading UART
					start_read_uart <= '0';
					-- start writting DAC
					start_write_dac <= '1';
					-- set UART_CTS HIGH to avoid reception of the next command signal from DTE
					uart_cts <= '1'; 
					state <= wrt_dac;
				else
					state <= rd_uart;
				end if;
			
			-- in this state FuncGen writes DAC, using the data received from DTE
			when wrt_dac =>
				uart_cts <= '1';
				start_write_uart <= '0';
				start_write_dac <= '1';
				start_read_uart <= '0';
				start_read_adc <= '0';
				-- DAC block keeps writting until write_done HIGH is asserted
				if (write_done_dac = '1') then
					-- stop writting DAC
					start_write_dac <= '0';
					-- start reading ADC (feedback)
					start_read_adc <= '1';
					state <= rd_adc;
				else
					state <= wrt_dac;
				end if;
			
			-- in this state FuncGen reads ADC
			when rd_adc =>
				uart_cts <= '1';
				start_write_uart <= '0';
				start_write_dac <= '0';
				start_read_uart <= '0';
				start_read_adc <= '1';
				if (read_done_adc = '1') then
					-- ADC keeps reading until read_done HIGH is asserted
					start_read_adc <= '0';
					state <= func;
				else
					state <= rd_adc;
				end if;
			
			-- in this state the 12-bit signal (from ADC) is converted
			-- into 16-bit word for transmission back to DTE via UART Tx
			when func =>
				uart_cts <= '1';
				start_write_uart <= '1';
				start_write_dac <= '0';
				start_read_uart <= '0';
				start_read_adc <= '0';
				-- MSB is STOP bit; next 4 bits are 0's; then 4 MSB's from ADC's signal;
				-- then START bit; STOP bit; then 8 lower bits from ADC's signal and START bit
				din_tx <= "10000"&dout_adc(11 downto 8)&"01"&dout_adc(7 downto 0)&'0';
				state <= wrt_uart;
			
			-- in this state FuncGen writes UART, sending back to DTE feedback signal
			when wrt_uart =>
				uart_cts <= '1';
				start_write_uart <= '1';
				start_write_dac <= '0';
				start_read_uart <= '0';
				start_read_adc <= '0';
				if (write_done_uart = '1') then
					-- Tx block keeps writing UART until write_done HIGH is asserted
					start_write_uart <= '0';
					-- and goes to idle state to finish the cycle
					state <= idle;
				else
					state <= wrt_uart;
				end if;
			
			end case;
		end if;
	end process fg;
	
end Behavioral;

